/***********************************************************************
 *
 * W D M D R V . C
 *
 * (C) 1997 Microsoft Corporation
 *
 *
 * Purpose:
 *  Simulate / Demonstrate the "real" WDM, binary compatible driver
 *  for the VxD mapping example.  The code in this driver reads the 
 *  button presses on a USB mouse.  
 *
 ***********************************************************************/ 


#define WANTVXDWRAPS
#define INITGUID

#include <wdm.h>
#include <hidusage.h>
#include <hidpi.h>
#include <hidclass.h>
#include <winerror.h>
#include <windef.h>
#include "wdmdrv.h"
#include "_wdmdrv.h"

#ifdef DEBUG
UINT dbgOpt=1;      // display full traces
#endif

// Local Prototypes 
UINT ProcessHIDData(IN PVOID Context);
VOID _cdecl DispatchNtReadFile(LPHID_MOUSE_DATA lpHidMouseData);
VOID ReadCompletion(IN PVOID Context, IN PIO_STATUS_BLOCK IoStatusBlock, IN ULONG Reserved);
VOID WorkItemCallback_Open( PVOID context );
VOID WorkItemCallback_Close( PVOID context );

NTSTATUS DriverEntry(
                        IN PDRIVER_OBJECT DriverObject,
                        IN PUNICODE_STRING RegistryPath
                    )
{
    //
    // Note that a real driver would setup the DriverObject, 
    // allocate data structures, etc. but in this simple example, this work
    // is not necessary.
    //

    DBGOUT(("==>DriverEntry"));
    DBGOUT(("<==DriverEntry"));
    return STATUS_SUCCESS;
}

/*
 *  UnicodeStringLength
 *
 */
ULONG UnicodeStringLength(PWCHAR str)
{
    ULONG result = 0;
    while (*str++)
    {
        result++;
    }
    return result;
}


/***********************************************************************
 *
 * OpenMouse
 *  
 * Purpose:
 *  Opens USB Mouse
 *  
 * Returns:
 *  HANDLE to open mouse, or NULL on error
 *
 * Notes:
 *  HANDLE is really pointer to HID_MOUSE_DATA structure.  Thus 
 *  Mouse can only be accessed through the ReadButton and CloseMouse
 *  functions 
 *
 *  This function is called from mapsys.sys
 *
 ***********************************************************************/
HANDLE DllExport _stdcall OpenMouse( )
{
    PWORK_QUEUE_ITEM    pWorkItemOpen;
    NTSTATUS            ntStatus;
    LPHID_MOUSE_DATA    lpHidMouseData;
    HANDLE              h = NULL;

    DBGOUT(( "==>OpenMouse" ));
    if (!(lpHidMouseData=
            (LPHID_MOUSE_DATA)ExAllocatePool(
                                              NonPagedPool,
                                              sizeof(HID_MOUSE_DATA)
                                            )))
    {
        DBGERR(("Alloc failed for HID_MOUSE_DATA"));
        return( NULL );
    }

    RtlZeroMemory( lpHidMouseData, sizeof(HID_MOUSE_DATA) );
#ifdef DEBUG
    lpHidMouseData->tag = MOUSE_DATA_TAG;   // mark data structure as valid for
#endif                                      // later asserts
    lpHidMouseData->bIsClosing = FALSE; 

    //
	// Initialize event object to wait for open to complete on worker thread
	//
	KeInitializeEvent( &(lpHidMouseData->eWaitForComplete), 
                        SynchronizationEvent, 
                        FALSE );

    //
    // Create work item to open mouse on system thread
    //
    if ( 
         !(
            pWorkItemOpen=
            (PWORK_QUEUE_ITEM)ExAllocatePool(NonPagedPool,
                                             sizeof(WORK_QUEUE_ITEM))
          )
       )
    {
        DBGERR(("Alloc failed for open work queue"));
        return( NULL );
    }
    ExInitializeWorkItem( pWorkItemOpen,
                          WorkItemCallback_Open, 
                          (PVOID)lpHidMouseData );
    
    //
    // pass in pointer to WORK_QUEUE_ITEM so it can be freed by
    // the work item function
    //
    lpHidMouseData->pWorkItemOpen = pWorkItemOpen;  

    ExQueueWorkItem( pWorkItemOpen, DelayedWorkQueue );

    //
    // Wait for delayed open to complete
    //
    do
    {
        ntStatus = KeWaitForSingleObject( 
                         &(lpHidMouseData->eWaitForComplete),
                         UserRequest,
                         UserMode,
                         TRUE,
                         NULL
                        );
    } while (STATUS_ALERTED == ntStatus);               
    
    if ( STATUS_SUCCESS != ntStatus )
    {
        DBGERR(( "Error %x waiting for open to complete", ntStatus ));
    }
    else if ( STATUS_SUCCESS != lpHidMouseData->LastStatus )
    {
        DBGERR(( "Detected Error %x returned from OpenUSBMouse", lpHidMouseData->LastStatus ));
    }
    else
    {
        // open went fine so kick of the first read

        // first set the operation complete event for this upcoming read to false
        KeClearEvent( &(lpHidMouseData->eWaitForComplete) );

        // now start the first read
        DispatchNtReadFile( lpHidMouseData );   

        h = (HANDLE) lpHidMouseData;    
    }
    DBGOUT(( "<==OpenMouse" ));
    return h;
}

/***********************************************************************
 *
 * ReadButtons
 *
 * Purpose:
 *  Reads the button status on USB mouse.
 *  
 * Parameters:
 *  IN HANDLE h // handle from the OpenMouse call (actually ptr to HID_MOUSE_DATA)
 *  OUT UINT pButtonStatus // 0 ==> no buttons (error), 1==>left button, 2==>right
 *  
 * Returns:
 *  NT_SUCCESS on success or appropriate error status
 *
 * Notes:
 *  Currently blocking reads are implemented (i.e. won't return until a button
 *  press is detected) by using the lpHidMouseData->eWaitForComplete Event Object.
 *  eWaitForComplete event is released by the ReadCompletion function when new
 *  data is detected.
 *
 *  This function is called from mapsys.sys
 *
 ***********************************************************************/ 
NTSTATUS DllExport _stdcall ReadButtons(IN HANDLE h, OUT UINT *pButtonStatus )
{
    NTSTATUS ntStatus;
    LPHID_MOUSE_DATA lpHidMouseData = (LPHID_MOUSE_DATA)h;

    DBGOUT(( "==>ReadButtons" ));

    if ( !h )
    {
        DBGERR(( "Null handle in ReadButtons" ));
        return STATUS_INVALID_PARAMETER;
    }
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));
    *pButtonStatus = 0;

    DBGOUT(("Waiting for read completion signal"));
    do
    {
        ntStatus = KeWaitForSingleObject( 
                             &(lpHidMouseData->eWaitForComplete),
                             UserRequest,
                             UserMode,
                             TRUE,
                             NULL
                            );
   } while (STATUS_ALERTED == ntStatus);                

   if ( !NT_SUCCESS( ntStatus ))
   {
       DBGERR (( "Wait for read completion failed" ));
       return ntStatus;
   }
   DBGOUT(("Got read completion signal"));
   *pButtonStatus = lpHidMouseData->buttonStatus;
   lpHidMouseData->buttonStatus = 0; 
   KeClearEvent( &(lpHidMouseData->eWaitForComplete) );

   DBGOUT(( "ReadButtons returning %d", *pButtonStatus ));
   DBGOUT(( "<==ReadButtons" ));
   return STATUS_SUCCESS;
}

/***********************************************************************
 *
 * CloseMouse
 *
 * Purpose:
 *      Shut downs mouse, by closing mouse device handle and deallocating
 *      memory
 *  
 * Parameters:
 *      HANDLE h;   // pointer to the HID_MOUSE_DATA structure for this  
 *                  // mouse    
 *  
 * Returns:
 *  STATUS_SUCCESS or appropriate error code on failure
 *
 * Notes:
 *   Called from mapsys.sys 
 *
 ***********************************************************************/
NTSTATUS DllExport _stdcall CloseMouse( HANDLE h )
{
    LPHID_MOUSE_DATA lpHidMouseData = (LPHID_MOUSE_DATA)h;

    DBGOUT(( "==>CloseMouse" ));

    if ( !h )
    {
        DBGERR(( "Null handle in ReadButtons" ));
        return STATUS_INVALID_PARAMETER;
    }

    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));
    if ( 
         !(
            lpHidMouseData->pWorkItemClose=
            (PWORK_QUEUE_ITEM)ExAllocatePool(NonPagedPool,
                                             sizeof(WORK_QUEUE_ITEM))
          )
       )
    {
        DBGERR(( "Error allocating close work item" ));
        return( STATUS_NO_MEMORY );
    }
    ExInitializeWorkItem( lpHidMouseData->pWorkItemClose,
                          WorkItemCallback_Close, 
                          (PVOID)lpHidMouseData );

    ExQueueWorkItem(lpHidMouseData->pWorkItemClose, DelayedWorkQueue);
    DBGOUT (("<==CloseMouse"));
}

/***********************************************************************
 *
 * WorkItemCallback_Close
 *
 * Purpose:
 *   Shutdown mouse by closing handle and freeing up resources.
 *  
 * Parameters:
 *   PVOID context // pointer to HID_MOUSE_DATA 
 *
 * Notes:
 *   It is necessary to do close operation on worker thread to allow
 *   any pending read to complete
 ***********************************************************************/ 

VOID WorkItemCallback_Close( PVOID context )
{
    LPHID_MOUSE_DATA  lpHidMouseData = (LPHID_MOUSE_DATA)context;
    NTSTATUS ntStatus;

    DBGOUT(( "==>WorkItemCallBack_Read " ));
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    //
    // Free memory for work item because it's not used again
    //
    ExFreePool( lpHidMouseData->pWorkItemClose );

    //
    // Tell the read loop {DispatchNtReadFile()==>WorkItemCallback_Read()==>
    // ReadCompletion()==>DispatchNtReadFile()} to stop 
    //
    KeClearEvent( &(lpHidMouseData->eWaitForComplete) );
    lpHidMouseData->bIsClosing = TRUE;  
    
    //
    // Wait for the read loop to terminate
    //
    do
    {
        ntStatus = KeWaitForSingleObject( 
                         &(lpHidMouseData->eWaitForComplete),
                         UserRequest,
                         UserMode,
                         TRUE,
                         NULL
                        );
    } while (STATUS_ALERTED == ntStatus);               
    
    if ( STATUS_SUCCESS != ntStatus )
    {
        DBGERR(( "Error %x waiting for read loop to terminate", ntStatus ));
        return;
    }

    //
    // At this point, the read loop has terminated so no background threads
    // will access the lpHidMouseData structure. Thus it's safe to free it.
    //
    if (lpHidMouseData->hidDescriptor)
    {
        ExFreePool(lpHidMouseData->hidDescriptor);
    }
    if (lpHidMouseData->report)
    {
        ExFreePool(lpHidMouseData->report);
    }
    if (lpHidMouseData->buttonValues)
    {
        ExFreePool(lpHidMouseData->buttonValues);
    }
    if (lpHidMouseData->valueCaps)
    {
        ExFreePool(lpHidMouseData->valueCaps);
    }
#ifdef DEBUG
    lpHidMouseData->tag = 0; // Mark this data structure as invalid
#endif
    ExFreePool( lpHidMouseData );
    ntStatus = ZwClose (lpHidMouseData->devHandle);
    if ( !NT_SUCCESS(ntStatus) )
    {
        DBGERR(("Error %x closing mouse handle %x", ntStatus, lpHidMouseData->devHandle));
    }
    DBGOUT(( "<==WorkItemCallBack_Close " ));
}
    
/***********************************************************************
 *
 * ReadCompletion
 *
 * Purpose:
 *  Callback function called from ZwReadFile function when read completes
 *  
 * Parameters:
 *   PVOID Context      // pointer to the correspoinding HID_MOUSE_DATA 
 *                      // structure for this read
 *   
 *   PIO_STATUS_BLOCK   // Io Status Block structure
 *   
 *   Reserved       
 *  
 * Returns:
 *  VOID
 *
 * Notes:
 *   Stores the result of the read, releases the eWaitForComplete 
 *   event when new data is obtained, and kicks off the next read
 *
 ***********************************************************************/ 
VOID ReadCompletion(IN PVOID Context, IN PIO_STATUS_BLOCK IoStatusBlock, IN ULONG Reserved)
{
    LPHID_MOUSE_DATA lpHidMouseData = (LPHID_MOUSE_DATA)Context;
    NTSTATUS Status = IoStatusBlock->Status;
    UINT u;

    DBGOUT(("==>ReadCompletion"));
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    if (STATUS_SUCCESS != Status )
    {
        if ( !NT_SUCCESS(Status) && (Status != STATUS_PENDING) )
        {
            DBGERR(("ZwReadFile error (Status=%xh)", (UINT)Status));            
        }
    }
    else
    {
        DBGOUT(( "Storing button status at %X", &(lpHidMouseData->buttonStatus) ));
        u = ProcessHIDData( lpHidMouseData );
        lpHidMouseData->buttonStatus |= u;

        //
        // If we detected a new button press, release any thread blocked in 
        // the _ReadButtons call to KeWaitForSingleEvent
        // 
        if ( u )    
        {
            KeSetEvent( &(lpHidMouseData->eWaitForComplete), 0, FALSE );
        }
    }

    //
    //  Start the next read if we're not in the process of shutting down
    //
    if ( !lpHidMouseData->bIsClosing)
    {
        DispatchNtReadFile( lpHidMouseData );   
    }
    else
    {
        //
        // If we're in the process of closing, release any thread blocked in
        // CloseMouse() because no new delayed reads will be issued, thus it
        // will be safe to free up all the data structures.
        //
        KeSetEvent( &(lpHidMouseData->eWaitForComplete), 0, FALSE );        
        DBGOUT(("Read loop terminating due to close request"));
    }
    DBGOUT(("<==ReadCompletion"));
}

/***********************************************************************
 *
 * ProcessHIDData
 *
 * Purpose:
 *  Helper function to ReadCompletion which decodes the data from
 *  the ZwReadFile call to indicate which buttons are pressed
 *  
 * Parameters:
 *  LPHID_MOUSE_DATA Pointer to the HID_MOUSE_DATA struture for this read
 *  
 * Returns:
 *  0 on error or no buttons pressed, 1 if left button pressed, or 2 
 *  right button is pressed. 
 *
 ***********************************************************************/
UINT ProcessHIDData(LPHID_MOUSE_DATA lpHidMouseData )
{
    PIO_STATUS_BLOCK IoStatusBlock = &(lpHidMouseData->ioStatusBlock);
    UINT numActualButtons;
    UINT uRet = 0;

    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    DBGOUT(("==>ProcessHIDData"));
    if (NT_SUCCESS(IoStatusBlock->Status))
    {
        BOOL scaledFailed = FALSE;

        DBGOUT(("Parsing mouse report: <%xh,%xh,%xh,%xh,%xh>",
                (UINT)lpHidMouseData->report[0],
                (UINT)lpHidMouseData->report[1],
                (UINT)lpHidMouseData->report[2],
                (UINT)lpHidMouseData->report[3],
                (UINT)lpHidMouseData->report[4]));

        /*
         *  Parse the mouse "report" for the values we want.
         */
        numActualButtons = lpHidMouseData->buttonListLength;
        HidP_GetButtons(HidP_Input,
                        HID_USAGE_PAGE_BUTTON,
                        0,  
                        lpHidMouseData->buttonValues,
                        &numActualButtons,
                        lpHidMouseData->hidDescriptor,
                        (PUCHAR)lpHidMouseData->report,
                        lpHidMouseData->hidCapabilities.InputReportByteLength);


        DBGOUT(("Mouse buttons %x", (UINT)lpHidMouseData->buttonValues[0] ));
        if ( (lpHidMouseData->buttonValues[0]>=1) && (lpHidMouseData->buttonValues[0]<=3) )
        {
            uRet = (UINT)lpHidMouseData->buttonValues[0];
        }

    }
    else
    {
        DBGERR(("ProcessHIDData returned error %x", (UINT)IoStatusBlock->Status));
    }

    DBGOUT(("<==ProcessHIDData"));  
    return( uRet );
}

/***********************************************************************
 *
 * DispatchNtReadFile
 *
 * Purpose:
 *  Dispatch the next read request to a worker thread
 *  
 * Parameters:
 *  LPHID_MOUSE_DATA lpHidMouseData     // pointer to the HidMouseData structure
 *  
 * Returns:
 *  VOID
 *
 *
 ***********************************************************************/
VOID _cdecl DispatchNtReadFile(LPHID_MOUSE_DATA lpHidMouseData)
{
        DBGOUT(("==> DispatchNtReadFile()"));
        /*
         *  Queue a work item to do the read; this way we'll be on a worker thread
         *  instead of (possibly) the NTKERN thread when we call NtReadFile().
         */
        ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));
        ExQueueWorkItem(&lpHidMouseData->workItemRead, DelayedWorkQueue);

        DBGOUT(("<== DispatchNtReadFile()"));
}


/***********************************************************************
 *
 * WorkItemCallback_Read
 *
 * Purpose:
 *  Perform the actual read on the USB mouse
 *  
 * Parameters:
 *  PVOID context   // pointer to the HID_MOUSE_DATA structure for this read
 *  
 * Returns:
 *  VOID
 *
 * Notes:
 *  Using non blocking reads to allow other threads to run.
 *  The ReadCompletion function is called system when read completes.
 *
 ***********************************************************************/
VOID WorkItemCallback_Read(PVOID context)
{
    LPHID_MOUSE_DATA lpHidMouseData = (LPHID_MOUSE_DATA) context;

    DBGOUT( ("==>[WorkItemCallback_Read] lpHdMouseData = %X", lpHidMouseData) );
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    //
    // do the read
    //
    ZwReadFile(
                lpHidMouseData->devHandle,
                NULL,
                ReadCompletion,
                lpHidMouseData,
                (PIO_STATUS_BLOCK)&lpHidMouseData->ioStatusBlock,
                (PVOID)lpHidMouseData->report,
                lpHidMouseData->hidCapabilities.InputReportByteLength,
                &lpHidMouseData->dataLength,
                NULL
              );
    DBGOUT( ("<==[WorkItemCallback_Read]"));
}

/***********************************************************************
 *
 * StoreHidMouseData
 *
 * Purpose:
 *  Helper function to the OpenUSBMouse function which stores various
 *  HID information about the mouse
 *  
 * Returns:
 *  STATUS_SUCCESS or STATUS_NO_MEMORY if allocation error occurs.
 *
 * Notes:
 *   Allocate memory for the HID_MOUSE_DATA structure.  This memory should
 *   be freed by ExFreePool at the time the mouse is closed.
 *
 ***********************************************************************/
NTSTATUS StoreHidMouseData( 
                        LPHID_MOUSE_DATA lpHidMouseData,
                        HANDLE devHandle,
                        PHIDP_CAPS caps,
                        PHIDP_PREPARSED_DATA desc,
                        UINT descSize,
                        PWCHAR deviceFileName)
{
    ULONG valueCapsLen;
    NTSTATUS ntStatus = STATUS_NO_MEMORY;

    DBGOUT(("==>StoreHidMouseData"));
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    lpHidMouseData->devHandle = devHandle;

    RtlCopyMemory(  (PVOID)&(lpHidMouseData->deviceFileName),
                    (PVOID)deviceFileName,
                    (UnicodeStringLength(deviceFileName)*sizeof(WCHAR))+sizeof(UNICODE_NULL));
    RtlCopyMemory((PVOID)&lpHidMouseData->hidCapabilities, (PVOID)caps, sizeof(HIDP_CAPS));

    /*
     *  Allocate space for the mouse descriptor.
     */
    if 
    ( 
        !(  lpHidMouseData->hidDescriptor = 
            (PHIDP_PREPARSED_DATA)ExAllocatePool(NonPagedPool, descSize)  )
    ) 
    {
        DBGERR(("Allocate for HID descriptor failed StoreHidMouseData"));
        goto StoreHidMouseDataError;
    }
    RtlCopyMemory((PVOID)lpHidMouseData->hidDescriptor, (PVOID)desc, descSize);

    /*
     *  Allocate space for the mouse report.
     */
    lpHidMouseData->report = (PUCHAR)ExAllocatePool(NonPagedPool,
                                                    lpHidMouseData->hidCapabilities.InputReportByteLength);
    if (!lpHidMouseData->report)
    {
        DBGERR(("_HeapAllocate for report buffer failed in StoreHidMouseData"));
        goto StoreHidMouseDataError;
    }

    /*
     *  Figure out the length of the buttons value
     *  and allocate a buffer for reading the buttons.
     */
    lpHidMouseData->buttonListLength = HidP_MaxUsageListLength(HidP_Input, HID_USAGE_PAGE_BUTTON, lpHidMouseData->hidDescriptor);
    DBGOUT(("Button values list length = %d.", lpHidMouseData->buttonListLength));
    if (lpHidMouseData->buttonListLength)
    {
        lpHidMouseData->buttonValues = (PUSAGE) ExAllocatePool(NonPagedPool,
                                                                   lpHidMouseData->buttonListLength * sizeof (USAGE));
        if (lpHidMouseData->buttonValues)
        {
            RtlZeroMemory(lpHidMouseData->buttonValues, lpHidMouseData->buttonListLength);
        }
        else
        {
            DBGERR(("HeapAlloc failed for button values buffer"));
            goto StoreHidMouseDataError;
        }
    }

    /*
     *  Allocate the array of value-caps.
     */
    valueCapsLen = caps->NumberInputValueCaps;
    if (valueCapsLen)
    {
        lpHidMouseData->valueCaps = (PHIDP_VALUE_CAPS)ExAllocatePool(NonPagedPool,
                                                                         valueCapsLen*sizeof(HIDP_VALUE_CAPS));
        if (!lpHidMouseData->valueCaps)
        {
            DBGERR(("HeapAlloc failed for value caps"));
            goto StoreHidMouseDataError;
        }
    }
    else
    {
        DBGERR(("value caps length = 0!"));
    }
    DBGOUT (("<==StoreHIDData"));
    return STATUS_SUCCESS;

StoreHidMouseDataError:
    if (lpHidMouseData->hidDescriptor)
    {
        ExFreePool(lpHidMouseData->hidDescriptor);
    }
    if (lpHidMouseData->report)
    {
        ExFreePool(lpHidMouseData->report);
    }
    if (lpHidMouseData->buttonValues)
    {
        ExFreePool(lpHidMouseData->buttonValues);
    }
    if (lpHidMouseData->valueCaps)
    {
        ExFreePool(lpHidMouseData->valueCaps);
    }
    ExFreePool(lpHidMouseData);
    DBGOUT (("<==StoreHIDData"));
    return ntStatus;
}

/***********************************************************************
 *
 * CallDeviceIoControl
 *
 * Purpose:
 *  Helper function to OpenUSBMouse that makes builds and sends IOCTRL
 *  IRPs to the HID drivers.
 *  
 * Returns:
 *  status from the IRP's status block
 *
 * Notes:
 *  Uses alertable waits and automatically retries wait if alerted.
 *
 ***********************************************************************/ 
NTSTATUS CallDeviceIoControl(   IN ULONG IoControlCode,
                                IN PDEVICE_OBJECT DeviceObject,
                                IN PVOID InputBuffer OPTIONAL,
                                IN ULONG InputBufferLength,
                                IN OUT PVOID OutputBuffer OPTIONAL,
                                IN ULONG OutputBufferLength,
                                OUT PIO_STATUS_BLOCK IoStatusBlock )
{
    NTSTATUS    ntStatus;
    KEVENT      IoCompleteEvent;
    PIRP        pIRP;

    DBGOUT(("==>CallDeviceIoControl"));
    KeInitializeEvent( &IoCompleteEvent, NotificationEvent, FALSE );
    pIRP = IoBuildDeviceIoControlRequest(
                                     IoControlCode,
                                     DeviceObject,
                                     InputBuffer,
                                     InputBufferLength,
                                     OutputBuffer,
                                     OutputBufferLength,
                                     FALSE,
                                     &IoCompleteEvent,
                                     IoStatusBlock);
    if ( !pIRP )
    {
        DBGERR(( "IoBuildIoControlRequestFailed" ));
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    ntStatus = IoCallDriver( DeviceObject, pIRP );

    if ( STATUS_PENDING == ntStatus )
    {
        do
        {
            ntStatus = KeWaitForSingleObject( 
                                             &IoCompleteEvent,
                                             Executive,
                                             KernelMode,
                                             TRUE,
                                             NULL
                                            );
        } while (STATUS_ALERTED == ntStatus);               

    }
    if ( !NT_SUCCESS( ntStatus ))
    {
        DBGERR (( "IoCallDriver failed" ));
        return ntStatus;
    }

    DBGOUT(("<==CallDeviceIoControl"));
    return pIRP->IoStatus.Status;
}


/***********************************************************************
 *
 * OpenUSBMouse
 *
 * Purpose:
 *  Do the actual work of opening the USB mouse
 *  
 * Parameters:
 *  LPHID_MOUSE_DATA lpHidMouseData // pointer to pointer of
 *                                  // main HID_MOSUE_DATA structure
 *                                  
 *  
 * Returns:
 *  STATUS_SUCCESS on success or appropriate error code on failur 
 *
 * Notes:
 *  Called from WorkItemCallback_open
 *
 ***********************************************************************/

NTSTATUS OpenUSBMouse( LPHID_MOUSE_DATA lpHidMouseData )
{
    IO_STATUS_BLOCK IoStatusBlock;
    NTSTATUS ntStatus = STATUS_UNSUCCESSFUL;
    OBJECT_ATTRIBUTES Obja;
    UNICODE_STRING FileName;
    PFILE_OBJECT pFileObject;
    PDEVICE_OBJECT pDO;
    PWSTR symbolicLinkList = NULL;
    PWSTR symbolicLink;
    ULONG entryCounter = 0;  
    BOOL  bOpenedMouse = FALSE; 

    /*
     *  Get a multi-string (separated by unicode NULL characters)
     *  of symbolic link names to the input-class devices.
     */
    DBGOUT(("==>OpenUSBMouse"));
    ntStatus = IoGetDeviceInterfaces(  (EXTERN_C GUID *)&GUID_CLASS_INPUT,
                                        NULL,
                                        0,
                                        (PWSTR *)&symbolicLinkList);
    if (NT_ERROR(ntStatus) || !symbolicLinkList)
    {
        DBGERR(("pIoGetDeviceInterfaces failed"));
        goto ErrOpenUSBMouse;
    }

    /*
     *  Go through all the device paths returned from GetDeviceInterfaces
     *  until we find a mouse.
     */ 
    
    symbolicLink = symbolicLinkList;
    while ( (WCHAR)*symbolicLink)
    {
        HANDLE deviceHandle;
        ULONG fileNameWChars;
        PWCHAR fileName;
    
        /*
         *  Get a pointer to to the next device name and step the multi-string pointer.
         */
        fileName = symbolicLink;
        fileNameWChars = UnicodeStringLength(fileName);
        symbolicLink += fileNameWChars+1;
    
        FileName.Buffer = fileName;
        FileName.Length = (USHORT)fileNameWChars*sizeof(WCHAR);
        FileName.MaximumLength = FileName.Length + sizeof(UNICODE_NULL);
    
        /*
         *  Initialize an object-attribute structure with this filename.
         */
        InitializeObjectAttributes(&Obja, &FileName, OBJ_CASE_INSENSITIVE, NULL, NULL);
    
        /*
         *  Try to open the device.
         */
        DBGOUT(("Opening HID Device : (unicode name @%xh, %xh wchars)", (UINT)fileName, (UINT)fileNameWChars));
    
        ntStatus = ZwCreateFile(
                                    &deviceHandle,
                                    (GENERIC_READ | SYNCHRONIZE | FILE_READ_ATTRIBUTES),
                                    &Obja,
                                    &IoStatusBlock,
                                    NULL,
                                    FILE_ATTRIBUTE_NORMAL,
                                    (FILE_SHARE_READ | FILE_SHARE_WRITE),
                                    FILE_OPEN,
                                    0,
                                    NULL,
                                    0
                                    );
        if (NT_SUCCESS(ntStatus))
        {
            HID_COLLECTION_INFORMATION hidColInfo;
    
            DBGOUT(("Opened some device (handle=%xh), calling _NtKernDeviceIoControl", (UINT)deviceHandle));

            ntStatus = IoGetDeviceObjectPointer(    
                                                    &FileName,
                                                    FILE_READ_DATA,
                                                    &pFileObject,
                                                    &pDO
                                               );
            if ( !NT_SUCCESS( ntStatus ) )
            {
                DBGERR(( "IoGetDeviceObjectPointer failed" ));
                goto ErrOpenUSBMouse;
            }

            ntStatus = CallDeviceIoControl(
                                           IOCTL_HID_GET_COLLECTION_INFORMATION,
                                           pDO,
                                           NULL,
                                           0,
                                           &hidColInfo,
                                           sizeof(HID_COLLECTION_INFORMATION),
                                           &IoStatusBlock);
 
            if (NT_SUCCESS(ntStatus))
            {
                PHIDP_PREPARSED_DATA phidDescriptor;
    
                phidDescriptor = (PHIDP_PREPARSED_DATA)ExAllocatePool(
                                                                      NonPagedPool,
                                                                      hidColInfo.DescriptorSize
                                                                     );
                if (phidDescriptor)
                {
                    ntStatus = CallDeviceIoControl(
                                                   IOCTL_HID_GET_COLLECTION_DESCRIPTOR,
                                                   pDO,
                                                   NULL,
                                                   0,
                                                   phidDescriptor,
                                                   hidColInfo.DescriptorSize,
                                                   &IoStatusBlock );
                    if (NT_SUCCESS(ntStatus))
                    {
                        HIDP_CAPS hidCaps;
    
                        ntStatus = HidP_GetCaps(phidDescriptor, (PHIDP_CAPS)&hidCaps);

    
                        if (NT_SUCCESS(ntStatus))
                        {
    
                            if (    (hidCaps.UsagePage == HID_USAGE_PAGE_GENERIC) &&
                                    ((hidCaps.Usage == HID_USAGE_GENERIC_MOUSE) || 
                                     (hidCaps.Usage == HID_USAGE_GENERIC_POINTER)))
                            {
    
                                DBGOUT(("Opened HID mouse successfully, report size is %d.", (UINT)hidCaps.InputReportByteLength));
    
                                if (hidCaps.InputReportByteLength == 0)
                                {
                                    DBGERR(("ERROR: Report size is zero!"));
                                }
                                else
                                {
                                    /*
                                     *  Take all the information we have for this mouse and bundle
                                     *  it into a context.
                                     */
                                    ntStatus = StoreHidMouseData (
                                                        lpHidMouseData, 
                                                        deviceHandle,
                                                        (PHIDP_CAPS)&hidCaps,
                                                        phidDescriptor,
                                                        hidColInfo.DescriptorSize,
                                                        fileName);
                                    if (NT_SUCCESS(ntStatus))
                                    {
                                        // setup background reads on this mouse
                                        ExInitializeWorkItem(&lpHidMouseData->workItemRead, 
                                                           WorkItemCallback_Read, 
                                                           lpHidMouseData);
                                        bOpenedMouse = TRUE;
                                        break;      // for this sample code, we will
                                                    // only open first mouse
                                    }
                                    else
                                    {
                                        DBGERR(("StoreHidMouseData() failed"));
                                    }
                                }
                            }
                            else
                            {
                                DBGERR(("The HID device's USAGE did not match that of a mouse (page=%xh,usage=%xh)", (UINT)hidCaps.UsagePage, (UINT)hidCaps.Usage));
                            }
                        }
                        else
                        {
                            DBGERR(("HidP_GetCaps failed"));
                        }
                    }
                    else
                    {
                        DBGERR(("DeviceIoControl (#2) failed"));
                    }
    
                    ExFreePool(phidDescriptor);
                    phidDescriptor = NULL;
                }
                else
                {
                    DBGERR(("ExAllocatePool failed"));
                }
            }
            else
            {
                DBGERR(("DeviceIoControl #1 failed"));
            }

        }
        else
        {
            DBGERR(("ZwCreateFile failed to open this Device (ntStatus=%xh)", (UINT)ntStatus));
        }
    }
    if ( !bOpenedMouse )
    {
        ntStatus = STATUS_NOT_FOUND;
    }
ErrOpenUSBMouse:

    if (symbolicLinkList)
    {
        ExFreePool(symbolicLinkList);
    }
    DBGOUT (("<==OpenUSBMouse"));
    return( ntStatus );
}

VOID WorkItemCallback_Open( PVOID context )
{
    LPHID_MOUSE_DATA  lpHidMouseData = (LPHID_MOUSE_DATA)context;

    DBGOUT(( "==>WorkItemCallBack_Open " ));
    ASSERT((MOUSE_DATA_TAG == lpHidMouseData->tag));

    //
    // Free memory for work item because it's not used again
    //
    ExFreePool( lpHidMouseData->pWorkItemOpen );

    lpHidMouseData->LastStatus = OpenUSBMouse( lpHidMouseData );

    //
    // Release thread waiting for open to complete
    // 
    KeSetEvent( &(lpHidMouseData->eWaitForComplete), 0, FALSE );
    
    DBGOUT (("<==WorkItemCallBack_Open"));
}

